#include <err.h>
#include <gmp.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "wrapper.h"

static const char *vecbeg = "[";
static const char *vecend = "]";
static const char *sep = ", ";
static const char *psep = "; ";

static void dimcoef(mpq_t, const ap_texpr0_t *, ap_dim_t, ap_environment_t *);
static int eval(mpq_t, const ap_texpr0_t *);
static void badconst(ap_environment_t *, ap_texpr0_t *, ap_texpr0_t *,
	ap_dim_t);
static const char *op_print(ap_texpr_op_t);

size_t
texpr_len(unsigned int flags, ap_environment_t *env, ap_texpr0_t *t,
	struct env *e)
{
	size_t l;
	ap_dim_t *d, *ds;
	mpq_t m;

	mpq_init(m);
	l = strlen(vecbeg);

	ds = ap_texpr0_dimlist(t);
	for(d = ds; *d != AP_DIM_MAX; d++) {
		dimcoef(m, t, *d, env);

		l += varlen(e, *d);
		l += mpq_len(m, flags);
		l += strlen(sep);

		if(d != ds)
			l += strlen(psep);
	}

	l += strlen(vecend);
	free(ds);
	mpq_clear(m);
	return l;
}

char *
texpr_pretty(char *s, unsigned int flags, ap_environment_t *env,
	ap_texpr0_t *t, struct env *e)
{
	ap_dim_t *d, *ds;
	mpq_t m;

	mpq_init(m);
	s = frag_pretty(s, vecbeg);
	ds = ap_texpr0_dimlist(t);
	for(d = ds; *d != AP_DIM_MAX; d++) {
		dimcoef(m, t, *d, env);

		if(d != ds)
			s = frag_pretty(s, psep);

		s = mpq_pretty(s, m, flags);
		s = frag_pretty(s, sep);
		s = varpretty(s, e, *d);
	}

	s = frag_pretty(s, vecend);
	free(ds);
	mpq_clear(m);
	return s;
}

void
texpr_cst(mpq_t r, const ap_texpr0_t *t)
{
	ap_texpr0_node_t *n;
	mpq_t ra, rb;

	if(!t)
		errx(1, "texpr_cst(): null t");

	switch(t->discr) {
	case AP_TEXPR_CST:
		coeff2mpq(r, &t->val.cst);
		return;

	case AP_TEXPR_DIM:
		mpq_set_ui(r, 0UL, 1UL);
		return;

	case AP_TEXPR_NODE:
		n = t->val.node;
		break;

	default:
		errx(1, "texpr_cst(): bad t discr");
	}

	if(ap_texpr_is_unop(n->op))
		switch(n->op) {
		case AP_TEXPR_NEG:
			texpr_cst(r, n->exprA);
			mpq_neg(r, r);
			return;

		default:
			errx(1, "texpr_cst(): unsupported unary op");
		}

	if(!ap_texpr_is_binop(n->op))
		errx(1, "texpr_cst(): bad op");

	mpq_init(ra);
	mpq_init(rb);
	texpr_cst(ra, n->exprA);
	texpr_cst(rb, n->exprB);

	switch(n->op) {
	case AP_TEXPR_ADD:
		mpq_add(r, ra, rb);
		break;

	case AP_TEXPR_SUB:
		mpq_sub(r, ra, rb);
		break;

	case AP_TEXPR_MUL:
		mpq_mul(r, ra, rb);
		break;

	case AP_TEXPR_DIV:
		if(mpq_sgn(rb) == 0)
			mpq_set_ui(r, 0UL, 1UL);
		else
			mpq_div(r, ra, rb);

		return;

	default:
		errx(1, "texpr_cst(): unsupported binary op");
	}

	mpq_clear(ra);
	mpq_clear(rb);
}

static void
dimcoef(mpq_t r, const ap_texpr0_t *t, ap_dim_t d, ap_environment_t *env)
{
	int za, zb;
	ap_texpr0_node_t *n;
	mpq_t ra, rb;

	if(!t)
		errx(1, "dimcoef(): null t");

	switch(t->discr) {
	case AP_TEXPR_CST:
		mpq_set_ui(r, 0UL, 1UL);
		return;

	case AP_TEXPR_DIM:
		mpq_set_ui(r, t->val.dim == d ? 1UL : 0UL, 1UL);
		return;

	case AP_TEXPR_NODE:
		n = t->val.node;
		break;

	default:
		errx(1, "dimcoef(): bad t discr");
	}

	if(ap_texpr_is_unop(n->op))
		switch(n->op) {
		case AP_TEXPR_NEG:
			dimcoef(r, n->exprA, d, env);
			mpq_neg(r, r);
			return;

		default:
			errx(1, "dimcoef(): unsupported unary op: %s", op_print(n->op));
		}

	if(!ap_texpr_is_binop(n->op))
		errx(1, "dimcoef(): bad op");

	mpq_init(ra);
	mpq_init(rb);
	dimcoef(ra, n->exprA, d, env);
	dimcoef(rb, n->exprB, d, env);
	za = mpq_sgn(ra) == 0;
	zb = mpq_sgn(rb) == 0;

	switch(n->op) {
	case AP_TEXPR_ADD:
		mpq_add(r, ra, rb);
		break;

	case AP_TEXPR_SUB:
		mpq_sub(r, ra, rb);
		break;

	case AP_TEXPR_MUL:
		if(!za && !zb)
			errx(1, "dimcoef(): non-linear");

		if(za && !zb)
			if(!eval(ra, n->exprA))
				badconst(env, n->exprA, n->exprB, d);

		if(zb && !za)
			if(!eval(rb, n->exprB))
				badconst(env, n->exprB, n->exprA, d);

		mpq_mul(r, ra, rb);
		break;

	case AP_TEXPR_DIV:
		if(za && zb) {
			mpq_set_ui(r, 0UL, 1UL);
			break;
		}

		if(!za && zb) {
			eval(rb, n->exprB);
			mpq_div(r, ra, rb);
			break;
		}

		errx(1, "dimcoef(): bad division");

	default:
		errx(1, "dimcoef(): unsupported binary op: %s", op_print(n->op));
	}

	mpq_clear(ra);
	mpq_clear(rb);
}

static int
eval(mpq_t r, const ap_texpr0_t *t)
{
	ap_texpr0_node_t *n;
	mpq_t ra, rb;

	if(!t)
		errx(1, "eval(): null t");

	switch(t->discr) {
	case AP_TEXPR_CST:
		coeff2mpq(r, &t->val.cst);
		return 1;

	case AP_TEXPR_NODE:
		n = t->val.node;
		break;

	default:
		return 0;
	}

	if(ap_texpr_is_unop(n->op))
		switch(n->op) {
		case AP_TEXPR_NEG:
			if(!eval(r, n->exprA))
				return 0;

			mpq_neg(r, r);
			return 1;

		default:
			errx(1, "eval(): unsupported unary op");
		}

	if(!ap_texpr_is_binop(n->op))
		errx(1, "eval(): bad operation");

	mpq_init(ra);
	mpq_init(rb);

	if(!eval(ra, n->exprA) || !eval(rb, n->exprB))
		return 0;

	switch(n->op) {
	case AP_TEXPR_ADD:
		mpq_add(r, ra, rb);
		break;

	case AP_TEXPR_SUB:
		mpq_sub(r, ra, rb);
		break;

	case AP_TEXPR_MUL:
		mpq_mul(r, ra, rb);
		break;

	case AP_TEXPR_DIV:
		mpq_div(r, ra, rb);
		break;

	default:
		mpq_clear(ra);
		mpq_clear(rb);
		errx(1, "eval(): unsupported binary op");
	}

	mpq_clear(ra);
	mpq_clear(rb);
	return 1;
}

static void
badconst(ap_environment_t *e, ap_texpr0_t *nbad, ap_texpr0_t *n, ap_dim_t d)
{
	char *v;
	ap_environment_name_of_dim_t *nd;

	nd = ap_environment_name_of_dim_alloc(e);
	v = ap_var_operations->to_string(e->var_of_dim[d]);

	ap_texpr0_fprint(stderr, nbad, nd->p);
	fputs(" is a bad coefficient for ", stderr);
	ap_texpr0_fprint(stderr, n, nd->p);
	fprintf(stderr, "\nVariable: %s\n", v);

	free((void *)v);
	ap_environment_name_of_dim_free(nd);
	errx(1, "badconst()");
}

static const char *
op_print(ap_texpr_op_t op)
{
	switch(op) {
	case AP_TEXPR_ADD: return "add";
	case AP_TEXPR_SUB: return "sub";
	case AP_TEXPR_MUL: return "mul";
	case AP_TEXPR_DIV: return "div";
	case AP_TEXPR_MOD: return "mod";
	case AP_TEXPR_NEG: return "neg";
	case AP_TEXPR_CAST: return "cast";
	case AP_TEXPR_SQRT: return "sqrt";
	default: return "unknown";
	}
}
