#include <err.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>

#include "wrapper.h"

enum {
	INCL = 0,
	CHULL = 1,
	ASSIGN = 2,
	MEET = 3,
	LAST = 4
};

static unsigned int init = 0;
static const unsigned int thresh = 1000;

static struct record recs[LAST] = {
	{ 0, 0, 0, NULL, NULL, "Incl" },
	{ 0, 0, 0, NULL, NULL, "Chull" },
	{ 0, 0, 0, NULL, NULL, "Assign" },
	{ 0, 0, 0, NULL, NULL, "Meet" }
};

static void startrec(struct record *, const char *);
static void stoprec(void);
static void finishf(struct record *);
static void startf(struct record *);
static FILE *fopenX(const char *);

bool __wrap_ap_abstract1_is_leq(ap_manager_t *m, ap_abstract1_t *a0,
		ap_abstract1_t *a1) {
	bool r;
	int rbeg, rend;
	double d;
	struct timespec beg, end;

	if(!check_manager(m) || !check_abstract(a0) || !check_abstract(a1))
		errx(1, "bad abstract domain");

	startrec(&recs[INCL], inclcall);

	rbeg = clock_gettime(CLOCK_THREAD_CPUTIME_ID, &beg);
	r = ap_abstract1_is_leq(m, a0, a1);
	rend = clock_gettime(CLOCK_THREAD_CPUTIME_ID, &end);

	if(rbeg != 0 || rend != 0)
		err(1, "clock_gettime(3)");

	d = tstod(&end) - tstod(&beg);
	inclrec(&recs[INCL], m, d, r, a0, a1);

	return r;
}

ap_abstract1_t __wrap_ap_abstract1_join(ap_manager_t *m, bool d,
		ap_abstract1_t *a1, ap_abstract1_t *a2) {
	int rbeg, rend;
	struct timespec beg, end;
	ap_abstract1_t a[2], r;

	if(!check_manager(m) || !check_abstract(a1) || !check_abstract(a2))
		errx(1, "bad abstract domain");

	if(d)
		errx(1, "__wrap_ap_abstract1_t()");

	startrec(&recs[CHULL], chullcall);

	rbeg = clock_gettime(CLOCK_THREAD_CPUTIME_ID, &beg);
	r = ap_abstract1_join(m, d, a1, a2);
	rend = clock_gettime(CLOCK_THREAD_CPUTIME_ID, &end);

	if(rbeg != 0 || rend != 0)
		err(1, "clock_gettime(3)");

	a[0] = *a1;
	a[1] = *a2;
	chullrec(&recs[CHULL], m, tstod(&end) - tstod(&beg), a, 2, &r);

	return r;
}

ap_abstract1_t __wrap_ap_abstract1_join_array(ap_manager_t *m,
		ap_abstract1_t *t, size_t sz) {
	int rbeg, rend;
	struct timespec beg, end;
	ap_abstract1_t r;
	unsigned int i;

	if(!check_manager(m))
		errx(1, "bad abstract domain");

	for(i = 0; i < sz; i++)
		check_abstract(&t[i]);

	startrec(&recs[CHULL], chullcall);

	rbeg = clock_gettime(CLOCK_THREAD_CPUTIME_ID, &beg);
	r = ap_abstract1_join_array(m, t, sz);
	rend = clock_gettime(CLOCK_THREAD_CPUTIME_ID, &end);

	if(rbeg != 0 || rend != 0)
		err(1, "clock_gettime(3)");

	chullrec(&recs[CHULL], m, tstod(&end) - tstod(&beg), t, sz, &r);

	return r;
}

ap_abstract1_t __wrap_ap_abstract1_assign_texpr_array(ap_manager_t *m,
	bool d, ap_abstract1_t *a, ap_var_t *tvar, ap_texpr1_t *te,
		size_t sz, ap_abstract1_t *dest) {
	int rbeg, rend;
	struct timespec beg, end;
	ap_abstract1_t r, b;

	if(!check_manager(m) || !check_abstract(a))
		errx(1, "bad abstract domain");

	startrec(&recs[ASSIGN], assigncall);

	b = ap_abstract1_copy(m, a);

	rbeg = clock_gettime(CLOCK_THREAD_CPUTIME_ID, &beg);
	r = ap_abstract1_assign_texpr_array(m, d, a, tvar, te, sz, dest);
	rend = clock_gettime(CLOCK_THREAD_CPUTIME_ID, &end);

	if(rbeg != 0 || rend != 0)
		err(1, "clock_gettime(3)");

	assignrec(&recs[ASSIGN], m, tstod(&end) - tstod(&beg), &b,
		tvar, te, sz, &r);

	ap_abstract1_clear(m, &b);
	return r;
}

ap_abstract1_t
__wrap_ap_abstract1_meet_tcons_array(ap_manager_t *m, bool d,
	ap_abstract1_t *a, ap_tcons1_array_t *t)
{
	int rbeg, rend;
	struct timespec beg, end;
	ap_abstract1_t r, b;

	if(!check_manager(m) || !check_abstract(a))
		errx(1, "bad abstract domain");

	startrec(&recs[MEET], meet_call);

	b = ap_abstract1_copy(m, a);

	rbeg = clock_gettime(CLOCK_THREAD_CPUTIME_ID, &beg);
	r = ap_abstract1_meet_tcons_array(m, d, a, t);
	rend = clock_gettime(CLOCK_THREAD_CPUTIME_ID, &end);

	if(rbeg != 0 || rend != 0)
		err(1, "clock_gettime(3)");

	meetrec(&recs[MEET], m, tstod(&end) - tstod(&beg), &b, t, &r);
	ap_abstract1_clear(m, &b);

	return r;
}

ap_abstract1_t
__wrap_ap_abstract1_meet(ap_manager_t *m, bool d, ap_abstract1_t *a1,
	ap_abstract1_t *a2)
{
	ap_abstract1_t res;
	ap_tcons1_array_t t;

	t = ap_abstract1_to_tcons_array(m, a2);
	res = __wrap_ap_abstract1_meet_tcons_array(m, d, a1, &t);
	ap_tcons1_array_clear(&t);

	return res;
}

static void startrec(struct record *r, const char *callstr) {
	if(!init) {
		init = 1;
		if(atexit(stoprec) != 0)
			err(1, "atexit(3)");
	}

	if(r->callstr == NULL)
		r->callstr = callstr;

	if(r->f == NULL)
		startf(r);
	else
		if(r->cnt - r->ffirst >= thresh) {
			finishf(r);
			startf(r);
		}
}

static void stoprec(void) {
	unsigned int i, j;
	char buf[50];
	const char *p;
	FILE *f, *fdep, *fmain;
	struct record *r;

	fmain = fopenX("Main.ml");

	fputs("let run () =\n", fmain);
	for(r = recs, i = 0; i < LAST; r++, i++)
		if(r->f != NULL) {
			p = r->fprefix;
			finishf(r);

			snprintf(buf, sizeof buf / sizeof buf[0], "%s.ml", p);
			f = fopenX(buf);

			fputs("let run () =\n", f);
			for(j = 0; j < r->fcnt; j++)
				fprintf(f, "\t%s%u.run ();\n", p, j);

			fclose(f);

			fprintf(fmain, "\t%s.run();\n", p);
		}

	fputs(";;\n\nrun ();\nStat.summary()\n", fmain);
	fclose(fmain);
}

static void startf(struct record *r) {
	char buf[50];

	snprintf(buf, sizeof buf / sizeof buf[0], "%s%u.ml",
		r->fprefix, r->fcnt++);
	r->f = fopenX(buf);
	r->ffirst = r->cnt;
}

static void finishf(struct record *r) {
	unsigned int i;

	fputs("let run () =\n", r->f);
	for(i = r->ffirst; i < r->cnt; i++)
		fprintf(r->f, r->callstr, i);

	fputs(";;\n", r->f);
	fclose(r->f);
	r->f = NULL;
}

static FILE *fopenX(const char *fname) {
	FILE *f;

	f = fopen(fname, "w");

	if(f == NULL)
		err(1, "fopen(3)");

	return f;
}
