(** The inclusion certificates computed by lipoly are manipulated using this module. *)

(** All the certificates aim at proving inclusion properties,
which can be proved by exhibiting linear combinations of constraints.
[frag_t] specifies such a linear combination of constraints as a list of pairs.
Each pair gives the coefficient to be applied to a constraint identified by a number
(see {! Cons.t}). *)
type frag_t = (int * Num.t) list

(** This type expresses the arguments of a proof of inclusion of a set of constraint in one constraint.
A proof can be [Direct], meaning that the provided linear combinations yields the desired contraint.
When proving inclusion in an equality, it might be useful to split it into two inequalities (see {! Cstr.split})
and provide a proof for each of them ([SplitEq]). *)
type cstr_t =
| Direct of frag_t
| SplitEq of frag_t * frag_t

(** This type expresses an overall strategy for proving the inclusion of [p1] in [p2]. *)
type t =
| Empty of frag_t (** Prove that [p1] is empty. *)
| Implies of (int * cstr_t) list (** Prove that [p1] is included in every constraint of [p2]. *)
| Bind of (int * Cstr.t * cstr_t * t) (** Introduce a new constraint as an intermediate step. *)

(** [isEqF f1 f2] returns [true] if [f1] and [f2] specify the same linear combination.
The order of the coefficients is irrelevant. *)
val isEqF: frag_t -> frag_t -> bool

(** Build a [frag_t] for a coefficient list.
[mkF] checks that none of the coefficients is nil. *)
(* XXX: explain why mkF should be preferred over manual construction. *)
val mkF: (int * Num.t) list -> frag_t

(** Apply a given function to all the coefficients of a [frag_t]. *)
val mapF: (Num.t -> Num.t) -> frag_t -> frag_t

(** Make linear combinations of fragments. *)
val mergeF: (frag_t * Num.t) list -> frag_t

(** [inlineF defs f] inlines the proofs for the constraints identified in [defs] when they appear in [f]. *)
val inlineF: (int * cstr_t) list -> frag_t -> frag_t

(** Pretty-print a linear combination. *)
val prF: frag_t -> string

(** [inlineC defs typ ce] inlines the proofs for the constraints identified in [defs] when they appear in [ce].
[typ] is the comparison operator of the constraint that [ce] refers to. *)
val inlineC: (int * cstr_t) list -> Cmp.t -> cstr_t -> cstr_t

(** Pretty-print a certificate for one constraint. *)
val prC: cstr_t -> string

(**/**)
(* XXX: remove *)
val mk: (int * cstr_t) list -> t

(* XXX: remove *)
val add: t -> int * cstr_t -> t
(**/**)

(** Pretty-print an inclusion certificate. *)
val pr: t -> string
