Require Import List.
Require NumC.
Require VecC.
Require ConsC.

Axiom id_t: Set.
Axiom eqId: id_t -> id_t -> bool.

Definition frag_t := list (id_t * NumC.t).

Inductive t: Set :=
| Direct: frag_t -> t
| SplitEq: frag_t -> frag_t -> t.

Definition Incl (l: list (id_t * ConsC.t)) (c: ConsC.t) :=
  forall x: VecC.t, Forall (fun c1: id_t * ConsC.t => ConsC.Sat (snd c1) x) l -> ConsC.Sat c x.

Lemma InclIn: forall (l: list (id_t * ConsC.t)) (c: id_t * ConsC.t), In c l -> Incl l (snd c).
  unfold Incl; intros l c I0 x F0;
  exact ((proj1 (Forall_forall (fun c1 => ConsC.Sat (snd c1) x) l)) F0 c I0).
Qed.

Lemma InclTrans: forall (l: list (id_t * ConsC.t)) (c1 c2: ConsC.t),
    Incl l c1 -> ConsC.Incl c1 c2 -> Incl l c2.
  unfold Incl, ConsC.Incl; intros l c1 c2 F1 F2 x F3;
  exact (F2 x (F1 x F3)).
Qed.

Lemma InclNil: forall (l: list (id_t * ConsC.t)), Incl l ConsC.nil.
  unfold Incl; intros; apply ConsC.NilSatSpec.
Qed.

Lemma InclAdd: forall (l: list (id_t * ConsC.t)) (c1 c2: ConsC.t),
    Incl l c1 -> Incl l c2 -> Incl l (ConsC.add c1 c2).
  unfold Incl; intros l c1 c2 I1 I2 x F0;
  exact (ConsC.AddSatSpec c1 c2 x (I1 x F0) (I2 x F0)).
Qed.

Lemma InclSplit: forall (l: list (id_t * ConsC.t))
  (c: { c0: ConsC.t | ConsC.typ c0 = Cmp.Eq }) (c1 c2: ConsC.t),
    fst (ConsC.splitEq c) = c1 -> snd (ConsC.splitEq c) = c2 ->
      Incl l c1 -> Incl l c2 -> Incl l (proj1_sig c).
  unfold Incl.
  intros l c c1 c2 Hc1 Hc2 Fc1 Fc2 x F.
  exact (ConsC.SplitEqSatSpec c c1 c2 Hc1 Hc2 x (Fc1 x F) (Fc2 x F)).
Qed.

Definition Contrad (l: list (id_t * ConsC.t)) :=
  forall x: VecC.t, not (Forall (fun c: id_t * ConsC.t => ConsC.Sat (snd c) x) l).

Lemma ContradIncl: forall (l: list (id_t * ConsC.t)) (c: ConsC.t),
    Contrad l -> Incl l c.
  unfold Contrad, Incl; intros l c F1 x F2; destruct (F1 x F2).
Qed.

Lemma ContradTrans: forall (l: list (id_t * ConsC.t)) (c: ConsC.t),
    Incl l c -> ConsC.Contrad c -> Contrad l.
  unfold Incl, ConsC.Contrad, Contrad;
  intros l c F1 F2 x F3;
  exact (F2 x (F1 x F3)).
Qed.

Fixpoint get (l: list (id_t * ConsC.t)) (n: id_t) { struct l }:
    Exc { c: id_t * ConsC.t | In c l }.
  refine (
  match l with
  | nil => error
  | (n1, c1)::l1 =>
    if eqId n n1 then
      value (exist _ (n1, c1) _)
    else
      match get l1 n with
      | value (exist c2 p2) => value _
      | error => error
      end
  end).
  exact (in_eq (n1, c1) l1).
  exact (exist _ c2 (in_cons (n1, c1) c2 l1 p2)).
Defined.

Definition mul (n: NumC.t) (c: ConsC.t): Exc { c1: ConsC.t | ConsC.Incl c c1 } :=
  match ConsC.typ c as t0 return
    ConsC.typ c = t0 -> Exc { c1: ConsC.t | ConsC.Incl c c1 }
  with
  | Cmp.Eq => (fun t0 =>
    let n1 := exist _ n (or_introl _ t0) in
    value (exist _ (ConsC.mul c n1) (ConsC.MulSatSpec c n1)))
  | _ => (fun _ =>
    match NumC.LtLeDec NumC.z n with
    | left lt1 =>
      let n1 := exist _ n (or_intror _ (NumC.LtNotLe NumC.z n lt1)) in
      value (exist _ (ConsC.mul c n1) (ConsC.MulSatSpec c n1))
    | right _ => error
    end)
  end (refl_equal (ConsC.typ c)).

Definition acc (l: list (id_t * ConsC.t)) (a: { c1: ConsC.t | Incl l c1 })
  (c: { c2: id_t * ConsC.t | In c2 l }) (n: NumC.t):
    Exc { c3: ConsC.t | Incl l c3 } :=
  let (c, pC) := c in
  let (a, pA) := a in
  match mul n (snd c) with
  | value (exist c4 p4) =>
    let p := InclAdd l _ _ pA (InclTrans l (snd c) c4 (InclIn l c pC) p4) in
    value (exist _ (ConsC.add a c4) p)
  | error => error
  end.

Fixpoint combine (l: list (id_t * ConsC.t)) (f: frag_t):
    Exc { c: ConsC.t | Incl l c } :=
  match f with
  | nil => value (exist _ ConsC.nil (InclNil l))
  | (n, coef)::f1 =>
    match combine l f1 with
    | value c1 =>
      match get l n with
      | value cn => acc l c1 cn coef
      | error => error
      end
    | error => error
    end
  end.

Definition checkIncl (l: list (id_t * ConsC.t)) (f: frag_t)
    (c: ConsC.t): Exc (Incl l c) :=
  match combine l f with
  | value (exist c1 p1) =>
    let i := ConsC.isIncl c1 c in
    match i as i0 return i = i0 -> Exc (Incl l c) with
    | true => (fun i1 =>
      let i2 := ConsC.IsInclSatSpec c1 c i1 in
      value (InclTrans l c1 c p1 i2))
    | false => (fun _ => error)
    end (refl_equal i)
  | error => error
  end.

Definition check (l: list (id_t * ConsC.t)) (ce: t)
   (c: ConsC.t): Exc (Incl l c).
  refine (
  match ce with
  | Direct f => checkIncl l f c
  | SplitEq f1 f2 =>
    match ConsC.typ c as t0 return
      ConsC.typ c = t0 -> Exc (Incl l c)
    with
    | Cmp.Eq => (fun t =>
      match ConsC.splitEq (exist _ c t) as _cSplit return
        ConsC.splitEq (exist _ c t) = _cSplit -> Exc (Incl l c)
      with
      | (c1, c2) => (fun cSplit =>
        match checkIncl l f2 c2 with
        | value p2 =>
          match checkIncl l f1 c1 with
          | value p1 => value (InclSplit l (exist _ c t) c1 c2 _ _ p1 p2)
          | error => error
          end
        | error => error
        end)
      end (refl_equal (ConsC.splitEq (exist _ c t))))
    | _ => (fun _ => error)
    end (refl_equal (ConsC.typ c))
  end);
  rewrite cSplit; reflexivity.
Defined.

Definition checkEmpty (l: list (id_t * ConsC.t)) (f: frag_t): Exc (Contrad l) :=
  match combine l f with
  | value (exist c1 p1) =>
    let i := ConsC.isContrad c1 in
    match i as i0 return i = i0 -> Exc (Contrad l) with
    | true =>
      (fun i1 => value (ContradTrans l c1 p1 (ConsC.IsContradSatSpec c1 i1)))
    | false => (fun _ => error )
    end (refl_equal i)
  | error => error
  end.
