Require Import Bool.
Require Import Ring.
Require NumC.

Add Ring NumRing: NumC.Ring.

Inductive t: Set :=
  | Nil: t
  | Sub: t -> NumC.t -> t -> t.

Fixpoint eval (v x: t): NumC.t :=
  match v, x with
  | Nil, _ => NumC.z
  | _, Nil => NumC.z
  | Sub l1 n1 r1, Sub l2 n2 r2 =>
    let l := eval l1 l2 in
    let r := eval r1 r2 in
    NumC.add (NumC.mul n1 n2) (NumC.add l r)
  end.

Definition Eq (v1 v2: t) := forall x: t, eval v1 x = eval v2 x.

Fixpoint normDep (v: t): sig (Eq v). (* specification is Eq here *)
  refine (
  match v as _v return sig (Eq _v) with
  | Nil => exist (Eq Nil) Nil (fun x: t => refl_equal (eval Nil x))
  | Sub l n r =>
    let (l1, lP) := normDep l in
    let (r1, rP) := normDep r in
    match l1 as _l1 return Eq l _l1 -> sig (Eq (Sub l n r)) with
    | Sub _ _ _  => (fun lP0 => exist _ (Sub l1 n r1) _)
    | Nil => (fun lP0 =>
      match r1 as _r1 return Eq r _r1 -> sig (Eq (Sub l n r)) with
      | Sub _ _ _  => (fun rP0 => exist _ (Sub l1 n r1) _)
      | Nil => (fun rP0 =>
        match NumC.EqDec NumC.z n with
        | left e => exist _ Nil _
        | right _ => exist _ (Sub l1 n r1) _
        end)
      end rP)
    end lP
  end);
  unfold Eq, eval; fold eval.
  destruct 0; [ | rewrite <- e, lP0, rP0, NumC.MulZL ]; reflexivity.
  destruct 0; [ | rewrite lP, rP ]; reflexivity.
  destruct 0; [ | rewrite lP, rP ]; reflexivity.
  destruct 0; [ | rewrite lP, rP ]; reflexivity.
Defined.

Definition norm (v: t): t := proj1_sig (normDep v).

Lemma NormSatSpec: forall v: t, Eq v (norm v).
  exact (fun v => proj2_sig (normDep v)).
Qed.

Fixpoint isEqSynDep (v1 v2: t): Exc (Eq v1 v2).
  refine
    (match v1 as _v1 return Exc (Eq _v1 v2) with
    | Nil =>
      match v2 as _v2 return Exc (Eq Nil _v2) with
      | Nil => value (fun _ => refl_equal _)
      | Sub _ _ _ => error
      end
    | Sub l1 n1 r1 =>
      match v2 as _v2 return Exc (Eq (Sub l1 n1 r1) _v2) with
      | Nil => error
      | Sub l2 n2 r2 =>
        match isEqSynDep l1 l2 with
        | value pl =>
          match isEqSynDep r1 r2 with
          | value pr =>
            match NumC.EqDec n1 n2 with
            | left pn => value _
            | right _ => error
            end
          | error => error
          end
        | error => error
        end
      end
    end).
  unfold Eq, eval; fold eval.
  destruct 0; [ | rewrite pl, pr, pn ]; reflexivity.
Defined.

Definition isEqDep (v1 v2: t): Exc (Eq v1 v2).
  refine (
    match isEqSynDep (norm v1) (norm v2) with
    | value e => value _
    | error => error
    end).
  unfold Eq; intro x.
  rewrite (NormSatSpec v1), (NormSatSpec v2).
  exact (e x).
Defined.

Definition isEq (v1 v2: t): bool :=
  match isEqDep v1 v2 with
  | value _ => true
  | error => false
  end.

Lemma IsEqSatSpec (v1 v2: t): isEq v1 v2 = true -> Eq v1 v2.
  unfold isEq; destruct (isEqDep v1 v2); intro h;
  [ assumption | inversion h ].
Qed.

Definition AddSpec (v1 v2 nv: t) :=
  forall x: t, eval nv x = NumC.add (eval v1 x) (eval v2 x).

Fixpoint addDep (v1 v2: t): sig (AddSpec v1 v2).
  refine (
  match v1 as _v1 return sig (AddSpec _v1 v2) with
  | Nil =>
    match v2 as _v2 return sig (AddSpec Nil _v2) with
    | Nil => exist (AddSpec Nil Nil) Nil (fun _ => refl_equal _)
    | Sub l n r => exist _ (Sub l n r) _
    end
  | Sub l1 n1 r1 =>
    match v2 as _v2 return sig (AddSpec (Sub l1 n1 r1) _v2) with
    | Nil => exist _ (Sub l1 n1 r1) _
    | Sub l2 n2 r2 =>
      let (nl, pl) := addDep l1 l2 in
      let (nr, pr) := addDep r1 r2 in
      exist _ (Sub nl (NumC.add n1 n2) nr) _
    end
  end);
  unfold AddSpec, eval; fold eval.
  destruct 0; [ | rewrite NumC.AddZL ]; reflexivity.
  destruct 0; [ | rewrite NumC.AddZR ]; reflexivity.
  destruct 0; [ reflexivity | rewrite pl, pr; ring ].
Defined.

Definition add (v1 v2: t): t := proj1_sig (addDep v1 v2).

Lemma AddSatSpec: forall (v1 v2: t), AddSpec v1 v2 (add v1 v2).
Proof (fun v1 v2 => proj2_sig (addDep v1 v2)).

Definition MulSpec (n: NumC.t) (v nv: t) :=
  forall x: t, eval nv x = NumC.mul n (eval v x).

Fixpoint mulDep (n: NumC.t) (v: t): sig (MulSpec n v).
  refine
  (match v as v0 return sig (MulSpec n v0) with
  | Nil => exist _ Nil _
  | Sub l n0 r =>
    let (l1, lP) := mulDep n l in
    let (r1, rP) := mulDep n r in
    exist _ (Sub l1 (NumC.mul n n0) r1) _
  end);
  [intro |]; unfold MulSpec, eval; fold eval;
  [| destruct 0 as [| l2 n2 r2]; [| rewrite (lP l2), (rP r2)]]; ring.
Defined.

Definition mul (n: NumC.t) (v: t): t := proj1_sig (mulDep n v).

Lemma MulSatSpec: forall (n: NumC.t) (v: t), MulSpec n v (mul n v).
Proof (fun n v => proj2_sig (mulDep n v)).

Definition NegSpec (v nv: t) := forall x: t, eval nv x = NumC.neg (eval v x).

Fixpoint negDep (v: t): sig (NegSpec v).
  refine (
  match v as v0 return sig (NegSpec v0) with
  | Nil => exist (NegSpec Nil) Nil (fun _ => refl_equal _)
  | Sub l n r =>
    let (l1, lP) := negDep l in
    let (r1, rP) := negDep r in
    exist _ (Sub l1 (NumC.neg n) r1) _
  end).
  unfold NegSpec, eval; fold eval; destruct 0;
  [ reflexivity | rewrite (lP x1), (rP x2); ring ].
Defined.

Definition neg (v: t): t := proj1_sig (negDep v).

Lemma NegSatSpec: forall (v: t), NegSpec v (neg v).
Proof (fun v => proj2_sig (negDep v)).

Lemma EvalNil: forall x: t, eval Nil x = NumC.z.
  reflexivity.
Qed.

Lemma AddZL: forall (v: t), add Nil v = v.
  destruct 0; reflexivity.
Qed.
