let samples: Db.t = List.fold_left Db.add Db.nil Data.db

module Key =
struct
	type t = { prg: string; op: Db.opT }

	let compare k1 k2 =
		let p = String.compare k1.prg k2.prg in
		if p <> 0 then p else Db.opCmp k1.op k2.op
end

module Stat = Map.Make ( Key )

type valueT = { apron: int64; libpoly: int64; check: int64;
	apronNT: int64; libpolyNT: int64; checkNT: int64 }
let valueZ: valueT = { apron = Int64.zero; libpoly = Int64.zero; check = Int64.zero;
	apronNT = Int64.zero; libpolyNT = Int64.zero; checkNT = Int64.zero }

type statT = valueT Stat.t
let statZ = Stat.empty

let update (o: Db.opT) (v: valueT) (s: Db.sampleT): valueT =
	let avg (l: int64 list): int64 =
		match l with [] -> failwith "F.update"
		| _::_ ->
			Int64.div (List.fold_left Int64.add Int64.zero l) (Int64.of_int (List.length l))
	in
	let sz = List.fold_left (+) 0 (s.Db.eqs @ s.Db.ineqs) in
	match o with
	| Db.Incl | Db.Join ->
		if sz > 1 then
			{ apron = Int64.add v.apron (avg s.Db.apron);
			libpoly = Int64.add v.libpoly (avg s.Db.libpoly);
			check = Int64.add v.check (avg s.Db.check);
			apronNT = Int64.add v.apronNT (avg s.Db.apron);
			libpolyNT = Int64.add v.libpolyNT (avg s.Db.libpoly);
			checkNT = Int64.add v.checkNT (avg s.Db.check) }
		else
			{ v with
			apron = Int64.add v.apron (avg s.Db.apron);
			libpoly = Int64.add v.libpoly (avg s.Db.libpoly);
			check = Int64.add v.check (avg s.Db.check) }
	| Db.Meet | Db.Assign ->
		if sz > 1 then
			{ v with
			apron = Int64.add v.apron (avg s.Db.apron);
			libpoly = Int64.add v.libpoly (avg s.Db.libpoly);
			apronNT = Int64.add v.apronNT (avg s.Db.apron);
			libpolyNT = Int64.add v.libpolyNT (avg s.Db.libpoly) }
		else
			{ v with
			apron = Int64.add v.apron (avg s.Db.apron);
			libpoly = Int64.add v.libpoly (avg s.Db.libpoly) }

let add (m: Db.metaT) (s: Db.sampleT) (stat: statT): statT =
	let k = { Key.prg = m.Db.prg; Key.op = m.Db.op } in
	let v = try Stat.find k stat with Not_found -> valueZ in
	Stat.add k (update k.Key.op v s) stat

let stat: statT = Db.M.fold add samples statZ

let prUnit n = 
	let u = ["ns"; "us"; "ms"; "s"] in
	let prD d64 =
		let d = Int64.to_int d64 in
		if d >= 100 then
			"." ^ (string_of_int d)
		else if d >= 10 then
			".0" ^ (string_of_int d)
		else if d > 0 then
			".00" ^ (string_of_int d)
		else
			""
	in
	let rec pr i d = function
	| [] -> (Int64.to_string i) ^ (prD d) ^ "min"
	| h::t ->
		let ni = Int64.div i (Int64.of_int 1000) in
		if Int64.compare Int64.zero ni = 0 then
			(Int64.to_string i) ^ (prD d) ^ h
		else
			let nd = Int64.rem i (Int64.of_int 1000) in
			pr ni nd t
	in
	pr n Int64.zero u

let statPrE (k: Key.t) (v: valueT): unit =
	Printf.printf "%s, %s: apron:%s (%s), libpoly:%s (%s), check: %s (%s)\n"
		k.Key.prg (Db.opPr k.Key.op) (prUnit v.apron) (prUnit v.apronNT)
		(prUnit v.libpoly) (prUnit v.libpolyNT) (prUnit v.check) (prUnit v.checkNT)
;;

Stat.iter statPrE stat

