let (x, b) = Basis.add Basis.nil "x"
let (y, b) = Basis.add b "y"

let mk l = List.map (fun (i, a) -> (i, Num.mk a)) l

let mkc t v c =
	Cstr.mk t (List.map (fun (c, v) -> (Num.mk c, v)) v) (Num.mk c)

let eq = mkc Cmp.Eq and le = mkc Cmp.Le and lt = mkc Cmp.Lt

let rec prPos = function
	| BinPos.Coq_xH -> "XH"
	| BinPos.Coq_xI p1 -> (prPos p1) ^ " XI"
	| BinPos.Coq_xO p1 -> (prPos p1) ^ " XO"

let prZ = function
	| BinInt.Z0 -> "Z0"
	| BinInt.Zpos p -> "Zpos " ^ (prPos p)
	| BinInt.Zneg p -> "Zneg " ^ (prPos p)

let prNum n =
	"num = " ^ (prZ n.QArith_base.coq_Qnum) ^
	"; den = " ^ (prPos n.QArith_base.coq_Qden)

(* Glue.zToPos *)
let zToPos_ok_ts =
	let chk (t, p, s) = OUnit.TestLabel (t, OUnit.TestCase (fun () ->
		let r = Glue.zToPos p in
		let e = "zToPos " ^ (Num.Z.pr p) ^ " = " ^ (prPos r) in
		OUnit.assert_bool e (r = s)))
	in
	let tcs = [
		"u", Num.Z.u, BinPos.Coq_xH;
		"two", Num.Z.mk 2, BinPos.Coq_xO BinPos.Coq_xH;
		"three", Num.Z.mk 3, BinPos.Coq_xI BinPos.Coq_xH;
		"four", Num.Z.mk 4, BinPos.Coq_xO (BinPos.Coq_xO BinPos.Coq_xH);
		"five", Num.Z.mk 5, BinPos.Coq_xI (BinPos.Coq_xO BinPos.Coq_xH);
		"six", Num.Z.mk 6, BinPos.Coq_xO (BinPos.Coq_xI BinPos.Coq_xH);
		"seven", Num.Z.mk 7, BinPos.Coq_xI (BinPos.Coq_xI BinPos.Coq_xH)
	] in
	OUnit.TestList (List.map chk tcs)

let zToPos_ko_ts =
	let chk (t, n) = OUnit.TestLabel (t, OUnit.TestCase (fun () ->
		OUnit.assert_raises (Invalid_argument "Glue.zToPos")
			(fun () -> Glue.zToPos n)))
	in
	let tcs = [
		"z", Num.Z.z;
		"negU", Num.Z.mk (-1);
		"minus_two", Num.Z.mk (-2)
	] in
	OUnit.TestList (List.map chk tcs)

let zToPos_ts = OUnit.TestList [
		OUnit.TestLabel ("ok", zToPos_ok_ts);
		OUnit.TestLabel ("ko", zToPos_ko_ts)
	]

(* Glue.zToZ *)
let zToZ_ts =
	let chk (t, z, s) = OUnit.TestLabel (t, OUnit.TestCase (fun () ->
		let r = Glue.zToZ z in
		let e = "zToZ " ^ (Num.Z.pr z) ^ " = " ^ (prZ r) in
		OUnit.assert_bool e (r = s)))
	in
	let tcs = [
		"z", Num.Z.z, BinInt.Z0;
		"u", Num.Z.u, BinInt.Zpos BinPos.Coq_xH;
		"two", Num.Z.mk 2, BinInt.Zpos (BinPos.Coq_xO BinPos.Coq_xH);
		"three", Num.Z.mk 3, BinInt.Zpos (BinPos.Coq_xI BinPos.Coq_xH);
		"negU", Num.Z.mk (-1), BinInt.Zneg BinPos.Coq_xH;
		"negTwo", Num.Z.mk (-2), BinInt.Zneg (BinPos.Coq_xO BinPos.Coq_xH);
	] in
	OUnit.TestList (List.map chk tcs)

(* Glue.nToNum *)
let nToNum_ts =
	let chk (t, n, s) = OUnit.TestLabel (t, OUnit.TestCase (fun () ->
		let r = Glue.nToNum n in
		let e = "nToNum " ^ (Num.pr n) ^ " = " ^ (prNum r) in
		OUnit.assert_bool e (r = s)))
	in
	let tcs = [
		"z", Num.z, { QArith_base.coq_Qnum = BinInt.Z0;
			QArith_base.coq_Qden = BinPos.Coq_xH };
		"u", Num.u, { QArith_base.coq_Qnum = BinInt.Zpos BinPos.Coq_xH;
			QArith_base.coq_Qden = BinPos.Coq_xH };
		"negU", Num.negU, { QArith_base.coq_Qnum = BinInt.Zneg BinPos.Coq_xH;
			QArith_base.coq_Qden = BinPos.Coq_xH }
	] in
	OUnit.TestList (List.map chk tcs)

(* Glue.check *)
let check_ok_ts: OUnit.test =
	let d c = Cert.Direct (Cert.mkF (mk c)) in
	let s c1 c2 = Cert.SplitEq (Cert.mkF (mk c1), Cert.mkF (mk c2))
	in
	let chk (t, l1, l2, c, r) = OUnit.TestLabel (t, OUnit.TestCase (fun () ->
		OUnit.assert_equal r (Glue.check l1 l2 (Cert.mk c))))
	in
	let tcs = [
		"simple0", [0, eq [1, x] 0], [0, eq [1, x] 0], [0, d [0, 1]], true;
		"two0", [0, eq [1, x] 0; 1, le [1, y] 0], [0, le [1, x; 1, y] 0],
			[0, d [0, 1; 1, 1]], true;

		"two1", [0, le [1, x] 0; 1, le [1, y] 0],
			[0, le [1, x; 2, y] 1; 1, le [2, x; 1, y] 0],
				[0, d [0, 1; 1, 2]; 1, d [0, 2; 1, 1]], true;

		"split0", [0, le [1, x] 0; 1, le [-1, x] 0], [0, eq [1, x] 0],
			[0, s [0, 1] [1, 1]], true;

		"fail0", [0, eq [1, x] 0; 1, le [1, y] 0], [0, le [1, x; 2, y] 0],
			[0, d [0, 1; 1, 1]], false;

		"missing0", [0, le [1, x] 0], [0, le [1, x] 0; 1, le [1, y] 0;
			2, le [1, x] 2], [0, d [0, 1]; 2, d [0, 1]; 3, d [0, 1]], false;

		"len0", [0, eq [1, x] 0], [0, eq [1, x] 0; 1, le [1, x] 0],
			[0, d [0, 1]], false;

		"len1", [0, eq [1, x] 0], [0, eq [1, x] 0],
			[0, d [0, 1]; 1, d [0, 1]], false
	] in
	OUnit.TestList (List.map chk tcs)

let check_bot_ts: OUnit.test =
	let chk (t, l, f, r) = OUnit.TestLabel (t, OUnit.TestCase (fun () ->
		OUnit.assert_equal r (Glue.check l [] (Cert.Empty (Cert.mkF (mk f))))))
	in
	let tcs = [
		"simple0", [0, le [] (-1)], [0, 1], true;
		"simple1", [0, lt [] 0], [0, 1], true;
		"simple2", [0, eq [] 1], [0, 1], true;
		"two0", [0, eq [1, x] 0; 1, eq [-1, x] 1], [0, 1; 1, 1], true;
		"two1", [0, le [1, x] 0; 1, le [-1, x] (-1)], [0, 1; 1, 1], true;
		"useless0", [0, le [] (-1); 1, le [] 0], [0, 1;  1, 1], true
	] in
	OUnit.TestList (List.map chk tcs)

let check_ts: OUnit.test = OUnit.TestList [
		OUnit.TestLabel ("ok", check_ok_ts);
		OUnit.TestLabel ("bot", check_bot_ts)
	]

(* Glue.checkMeet *)
let checkMeet_ts =
	let d c = Cert.Direct (Cert.mkF (mk c)) in
	let s c1 c2 = Cert.SplitEq (Cert.mkF (mk c1), Cert.mkF (mk c2))
	in
	let chk (t, r, c, l1, l2, ce) = OUnit.TestLabel (t, OUnit.TestCase (fun () ->
		OUnit.assert_equal r (Glue.checkMeet l1 c l2 ce)))
	in
	let tcs = [
		"simple0", true, (1, le [1, x] 0), [0, le [1, y] 0],
			Some [0, le [1, y] 0; 1, le [1, x] 0],
			Cert.Implies [0, d [0, 1]; 1, d [1, 1]];

		(* check that order is cared for. *)
		"simple1", true, (1, le [1, x] 0), [0, le [1, y] 0],
			Some [0, le [1, y] 0; 1, le [1, x] 0],
			Cert.Implies [1, d [1, 1]; 0, d [0, 1]];

		"bind0", true, (3, le [1, x] 0), [0, le [-1, x] 0; 1, le [1, x; 1, y] 0],
			Some [0, eq [1, x] 0; 1, le[1, y] 0],
			Cert.Bind (4, eq [1, x] 0, s [3, 1] [0, 1], Cert.Implies [0, d [4, 1]; 1, d [1, 1; 4, -1]])
	] in
	OUnit.TestList (List.map chk tcs)

let ts = OUnit.TestList [
		OUnit.TestLabel ("zToPos", zToPos_ts);
		OUnit.TestLabel ("zToZ", zToZ_ts);
		OUnit.TestLabel ("nToNum", nToNum_ts);
		OUnit.TestLabel ("check", check_ts);
		OUnit.TestLabel ("checkMeet", checkMeet_ts)
	]
